from .base import API21Object, Date
from .company import Contact
from decimal import Decimal
from pydantic import BaseModel
from typing import List, Optional


class SOTransactionItem(BaseModel):
    bundlenumber: Optional[str] = None
    itemid: str = ...
    itemdesc: Optional[str] = None
    taxable: Optional[bool] = None
    warehouseid: Optional[str] = None
    quantity: Decimal = ...
    unit: str = ...
    linelevelsimpletax: Optional[str] = None
    discountpercent: Optional[Decimal] = None
    price: Decimal = ...
    sourcelinekey: Optional[int] = None
    discsurchargememo: Optional[str] = None
    locationid: Optional[str] = None
    departmentid: Optional[str] = None
    memo: Optional[str] = None
    # itemdetails
    # customfields
    revrectemplate: Optional[str] = None
    revrecstartdate: Optional[Date] = None
    revrecenddate: Optional[Date] = None
    renewalmacro: Optional[str] = None
    projectid: Optional[str] = None
    customerid: Optional[str] = None
    vendorid: Optional[str] = None
    employeeid: Optional[str] = None
    classid: Optional[str] = None
    contractid: Optional[str] = None
    # fulfillmentstatus
    taskno: Optional[str] = None
    billingtemplate: Optional[str] = None
    dropship: Optional[bool] = None
    shipto: Optional[Contact] = None


class SOTransactionItems(BaseModel):
    sotransitem: List[SOTransactionItem]


class SOTransaction(API21Object):
    transactiontype: str = ...
    datecreated: Date = ...
    dateposted: Optional[Date] = None
    createdfrom: Optional[str] = None
    customerid: str = ...
    documentno: Optional[str] = None
    origdocdate: Optional[Date] = None
    referenceno: Optional[str] = None
    termname: Optional[str] = None
    datedue: Optional[Date] = None
    message: Optional[str] = None
    shippingmethod: Optional[str] = None
    shipto: Optional[Contact] = None
    billto: Optional[Contact] = None
    supdocid: Optional[str] = None
    externalid: Optional[str] = None
    basecurr: Optional[str] = None
    currency: Optional[str] = None
    exchratedate: Optional[Date] = None
    exchratetype: Optional[str] = None
    exchrate: Optional[str] = None
    vsoepricelist: Optional[str] = None
    # customfields
    state: Optional[str] = None
    projectid: Optional[str] = None
    sotransitems: SOTransactionItems = ...
    # subtotals



from .base import API21Object, Date
from .company import Contact
from decimal import Decimal
from pydantic import BaseModel
from typing import List


class POTransactionItem(BaseModel):
    itemid: str = ...
    itemdesc: str = None
    taxable: bool = None
    warehouseid: str = None
    quantity: Decimal = ...
    unit: str = None
    linelevelsimpletaxtype: str = None
    price: Decimal = None
    sourcelinekey: int = None
    overridetaxamount: Decimal = None
    deliverto: Contact = None
    tax: Decimal = None
    locationid: str = None
    departmentid: str = None
    memo: str = None
    # itemdetails
    form1099: bool = None
    form1099type: str = None
    form1099box: str = None
    # customfields
    projectid: str = None
    customerid: str = None
    vendorid: str = None
    employeeid: str = None
    classid: str = None
    contractid: str = None
    billable: bool = None


class POTransactionItems(BaseModel):
    potransitem: List[POTransactionItem]


class POTransaction(API21Object):
    transactiontype: str = None
    datecreated: Date = None
    dateposted: Date = None
    createdfrom: str = None
    vendorid: str = None
    documentno: str = None
    referenceno: str = None
    vendordocno: str = None
    termname: str = None
    datedue: Date = None
    message: str = None
    shippingmethod: str = None
    returnto: Contact = None
    payto: Contact = None
    deliverto: Contact = None
    supdocid: str = None
    externalid: str = None
    basecurr: str = None
    currency: str = None
    exchratedate: Date = None
    exchratetype: str = None
    exchrate: Decimal = None
    # customfields
    state: str = None
    potransitems: POTransactionItems = None
    # subtotals



from pydantic import BaseModel


class Date(BaseModel):
    year: str
    month: str
    day: str


class API21Object(BaseModel):
    """A base class for implementing API 2.1 classes."""
    @classmethod
    def create(cls) -> str:
        """The function name for creating an object of this type."""
        return f'create_{cls.__name__.lower()}'

    @classmethod
    def delete(cls) -> tuple:
        """A tuple of the function name for deleting the object
        type and the name of the XML attribute used as the key."""
        return f'delete_{cls.__name__.lower()}', 'key'



from pyintacct.models.base import API21Object, Date
from decimal import Decimal
from pydantic import BaseModel
from typing import List


class LineItemDetail(BaseModel):
    glaccountno: str
    amount: Decimal
    memo: str = None
    locationid: str = None
    departmentid: str = None
    customerid: str = None
    vendorid: str = None
    employeeid: str = None
    itemid: str = None
    classid: str = None
    contractid: str = None
    warehouseid: str = None


class LineItem(BaseModel):
    lineitem: List[LineItemDetail]


class ARInvoice(API21Object):
    customerid: str = ...
    datecreated: Date = ...
    dateposted: Date = None
    datedue: Date = None
    termname: str = None
    action: str = ...
    invoiceno: str = None
    ponumber: str = None
    description: str = None
    basecurr: str = None
    currency: str = None
    exchratetype: str = None
    invoiceitems: LineItem = ...

    @classmethod
    def create(cls) -> str:
        return 'create_invoice'

    @classmethod
    def delete(cls):
        return 'delete_invoice', 'key'



import logging
import requests
import time
from copy import deepcopy
from jxmlease import parse, XMLDictNode, XMLCDATANode
from pydantic import BaseModel
from typing import List, Type, Union
from uuid import uuid4
from .exceptions import IntacctException, IntacctServerError
from .models.base import API21Object


logging.basicConfig(level=logging.INFO, format='%(asctime)s:%(levelname)s:%(message)s')

BASE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<request>
    <control>
        <senderid></senderid>
        <password></password>
        <controlid></controlid>
        <uniqueid></uniqueid>
        <dtdversion></dtdversion>
        <includewhitespace></includewhitespace>
    </control>
    <operation>
        <authentication></authentication>
        <content></content>
    </operation>
</request>
"""


class IntacctAPI(object):
    def __init__(self, config: dict = None):
        self.config = {
            'SENDER_ID': None,
            'SENDER_PW': None,
            'COMPANY_ID': None,
            'ENTITY_ID': None,
            'USER_ID': None,
            'USER_PW': None,
            'SESSION_ID': None,
            'ENDPOINT': 'https://api.intacct.com',
            'UNIQUE_ID': 'false',
            'DTD_VERSION': '3.0',
            'INCLUDE_WHITESPACE': 'true'
        }
        if config is not None:
            self.config.update(config)
        self.debug_level = logging.DEBUG
        self.session = requests.session()
        self.session.headers = {'content-type': 'application/xml',
                                'accept-encoding': '*',
                                'user-agent': 'pyintacct-0.0.7'}
        self.url = 'https://api.intacct.com/ia/xml/xmlgw.phtml'
        self.basexml = parse(BASE_XML)
        self.basexml['request']['control'].update(XMLDictNode({
            'senderid': self.config['SENDER_ID'],
            'password': self.config['SENDER_PW'],
            'controlid': uuid4(),
            'uniqueid': self.config['UNIQUE_ID'],
            'dtdversion': self.config['DTD_VERSION'],
            'includewhitespace': self.config['INCLUDE_WHITESPACE'],
        }))
        self.sessionid = None
        self.session_expiration = 0

    def execute(self, payload: XMLDictNode, refresh_session=True) -> XMLDictNode:
        """
        Sends the request to the Intacct API. Automatically refreshes session token after one hour.

        :param payload: A jxmlease structure containing the full payload except authentication.
        :param refresh_session: An override flag to allow bypassing session reuse.
        :return: A requests response.
        """
        if self.session_expiration < time.time() and refresh_session:
            try:
                self.sessionid = self.get_session_id()
                # TODO: The expiration can be obtained from the API response now.
                self.session_expiration = time.time() + 60 * 60
            except IntacctException as e:
                self.sessionid = None
                self.session_expiration = 0
        if refresh_session:
            payload['request']['operation']['authentication'].add_node(tag='sessionid',
                                                                       new_node=XMLCDATANode(self.sessionid))
        try:
            payload.standardize()
            r = self.session.post(self.url, data=payload.emit_xml().encode('utf-8'))
            if 500 <= r.status_code <= 599:
                # If a 500 error is encountered we raise IntacctServerError. The user may decide whether to retry.
                raise IntacctServerError(r.text)
            response = parse(r.text)
            if self.validate_response(response):
                return response
            else:
                raise IntacctException('Intacct API call failed.\n' + r.text)
        except requests.exceptions.RequestException as e:
            raise IntacctException(e)

    def get_function_base(self):
        """
        Internal function to get a prepared copy of the XML request and the function node.
        :return: Tuple of the complete payload and the function node.
        """
        payload = deepcopy(self.basexml)
        content = payload['request']['operation']['content']
        function = content.add_node(tag='function', new_node=XMLDictNode(tag='function'))
        function.set_xml_attr('controlid', str(uuid4()))
        return payload, function

    @staticmethod
    def validate_response(xml: XMLDictNode) -> bool:
        """

        :param xml: jxmlease XML tree to validate.
        :return: Returns True if the tree has no error nodes,
                 otherwise raises an IntacctException with the error node contents.
        """
        msg = ''
        for node in xml.find_nodes_with_tag('error'):
            msg += f'{node.get("description")}{node.get("description2")}\n'
        if msg != '':
            raise IntacctException(msg)
        else:
            return True

    def get_session_id(self) -> str:
        payload = deepcopy(self.basexml)
        login = XMLDictNode({
            'userid': self.config['USER_ID'],
            'companyid': self.config['COMPANY_ID'],
            'password': self.config['USER_PW']
        })
        payload['request']['operation']['authentication'].add_node(tag='login', new_node=login)
        function = XMLDictNode(tag='function')
        function.set_xml_attr('controlid', str(uuid4()))
        function.add_node('getAPISession')
        payload['request']['operation']['content'].add_node(tag='function', new_node=function)
        response = self.execute(payload, refresh_session=False)
        sessionid = next(response.find_nodes_with_tag('sessionid'))
        return str(sessionid.text)

    def read_by_query(self, obj: str, query: str, fields: str = '*', pagesize: int = 100, docparid: str = ''):
        """TODO: support automatic deserialisation instead of returning jxmlease object
           TODO: turn into generator instead of all results at once.
        """
        payload, function = self.get_function_base()
        function.add_node(tag='readByQuery', new_node=XMLDictNode({
            'object': obj,
            'fields': fields,
            'query': query,
            'pagesize': pagesize,
            'docparid': docparid}))
        results = list()
        r = self.execute(payload)
        results.extend(r.find_nodes_with_tag(obj.lower()))
        data = next(r.find_nodes_with_tag('data'))
        remaining = data.get_xml_attr('numremaining')
        result_id = data.get_xml_attr('resultId')
        while int(remaining) > 0:
            data, remaining, result_id = self.read_more(result_id)
            results.extend(data.find_nodes_with_tag(obj.lower()))
        return results

    def read_more(self, result_id):
        payload, function = self.get_function_base()
        function.add_node(tag='readMore', new_node=XMLDictNode({
            'resultId': result_id
        }))
        r = self.execute(payload)
        data = next(r.find_nodes_with_tag('data'))
        remaining = data.get_xml_attr('numremaining')
        result_id = data.get_xml_attr('resultId', None)
        return data, remaining, result_id
        
    def inspect(self, obj: str = '*', detail: bool = False, name: str = None):
        payload, function = self.get_function_base()
        inspect_node = function.add_node(tag='inspect', new_node=XMLDictNode({
            f'{"name" if name else "object"}': f'{name if name else obj}'}
        ))
        inspect_node.set_xml_attr('detail', f'{"1" if detail else "0"}')
        return self.execute(payload)

    def create(self, obj):
        payload, function = self.get_function_base()
        tag = 'create'
        if issubclass(obj.__class__, API21Object):
            tag = obj.create()
            new_node = XMLDictNode(obj.dict(skip_defaults=True))
        elif hasattr(obj, 'dict'):
            obj = dict([(obj.__class__.__name__.upper(), obj.dict())])
            new_node = XMLDictNode(obj)
        else:
            new_node = XMLDictNode(obj)
        function.add_node(tag=tag, new_node=new_node)
        return self.execute(payload)

    def update(self, obj: Union[dict, BaseModel]):
        payload, function = self.get_function_base()
        tag = 'update'
        if hasattr(obj, 'dict'):
            obj = dict([(obj.__class__.__name__.upper(), obj.dict())])
            new_node = XMLDictNode(obj)
        else:
            new_node = XMLDictNode(obj)
        function.add_node(tag=tag, new_node=new_node)
        return self.execute(payload)

    def _delete_v21(self, obj: type(API21Object), keys: List[str]):
        function_name, key_attr = obj.delete()
        for key in keys:
            payload, function = self.get_function_base()
            f = function.add_node(function_name)
            f.set_xml_attr(key_attr, key)
            self.execute(payload)
        return

    def delete(self, obj: Union[str, type(API21Object)], keys: List[str]):
        try:
            if issubclass(obj, API21Object):
                return self._delete_v21(obj, keys)
        except TypeError:
            pass
        payload, function = self.get_function_base()
        function.add_node(tag='delete', new_node=XMLDictNode({
            'object': obj,
            'keys': ','.join(keys)
        }))
        return self.execute(payload)



from setuptools import find_packages, setup


NAME = 'pyintacct'
VERSION = '0.0.7'
DESCRIPTION = 'Python SDK for Intacct'
LONG_DESCRIPTION = open('README.md').read()
LICENSE = 'MIT'
AUTHOR = 'red-coracle'
AUTHOR_EMAIL = ''
URL = 'https://github.com/red-coracle/pyintacct'
PYTHON_VERSION = '>=3.6.0'
REQUIRES = ['requests >=2.22, <3.0',
            'jxmlease >= 1.0.1',
            'pydantic == 0.32.2']



setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    python_requires=PYTHON_VERSION,
    install_requires=REQUIRES,
    # test_suite='tests',
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7']
)



from .base import API21Object
from pydantic import BaseModel
from typing import Optional


class MailAddress(BaseModel):
    address1: Optional[str] = None
    address2: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    zip: Optional[str] = None
    country: Optional[str] = None
    isocountrycode: Optional[str] = None
    latitude: Optional[str] = None
    longitude: Optional[str] = None


class Contact(API21Object):
    contactname: str = ...
    printas: Optional[str] = None
    companyname: Optional[str] = None
    taxable: Optional[bool] = None
    taxgroup: Optional[str] = None
    prefix: Optional[str] = None
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    initial: Optional[str] = None
    phone1: Optional[str] = None
    phone2: Optional[str] = None
    cellphone: Optional[str] = None
    pager: Optional[str] = None
    fax: Optional[str] = None
    email1: Optional[str] = None
    email2: Optional[str] = None
    url1: Optional[str] = None
    url2: Optional[str] = None
    status: Optional[str] = None
    mailaddress: Optional[MailAddress] = None
    taxid: Optional[str] = None

    @classmethod
    def delete(cls):
        return 'delete_contact', 'contactname'
